use clap::Parser;
use std::str::FromStr;
use std::time::Duration;
use tokio::io::{AsyncReadExt, AsyncWriteExt};
use tokio::net::{TcpStream, UdpSocket};

// Custom copy function that flushes after each write to avoid buffering
async fn copy_with_flush<R, W>(reader: &mut R, writer: &mut W) -> tokio::io::Result<u64>
where
    R: AsyncReadExt + Unpin,
    W: AsyncWriteExt + Unpin,
{
    let mut buf = [0u8; 1024]; // Use a small buffer for low latency
    let mut total_copied = 0u64;

    loop {
        match reader.read(&mut buf).await {
            Ok(0) => break, // EOF
            Ok(n) => {
                writer.write_all(&buf[..n]).await?;
                writer.flush().await?; // Flush immediately to avoid buffering
                total_copied += n as u64;
            }
            Err(e) => return Err(e),
        }
    }

    Ok(total_copied)
}

#[derive(Debug, Clone)]
enum Protocol {
    Tcp,
    Udp,
}

#[derive(Debug, Clone)]
struct KnockTarget {
    port: u16,
    protocol: Protocol,
}

impl FromStr for KnockTarget {
    type Err = String;

    fn from_str(s: &str) -> Result<Self, Self::Err> {
        let parts: Vec<&str> = s.split('/').collect();
        if parts.len() != 2 {
            return Err("Invalid format. Expected PORT/PROTOCOL".to_string());
        }

        let port = parts[0]
            .parse::<u16>()
            .map_err(|e| format!("Invalid port number: {}", e))?;
        let protocol = match parts[1].to_lowercase().as_str() {
            "tcp" => Protocol::Tcp,
            "udp" => Protocol::Udp,
            _ => return Err("Invalid protocol. Use 'tcp' or 'udp'".to_string()),
        };

        Ok(KnockTarget { port, protocol })
    }
}

#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Cli {
    /// The target to connect to after knocking, e.g., 127.0.0.1:22
    #[arg(required = true)]
    target_address: String,

    /// Port knocking sequence, e.g., --knock 1000/tcp --knock 2000/udp
    #[arg(long, short, value_parser = clap::value_parser!(KnockTarget), required = true, env = "PORT_KNOCKING_SEQ", value_delimiter = ':', num_args = 1..)]
    knock: Vec<KnockTarget>,
}

#[tokio::main]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    let cli = Cli::parse();

    tracing_subscriber::fmt().with_writer(std::io::stderr).init();

    let (target_host, _target_port) = {
        let parts: Vec<&str> = cli.target_address.split(':').collect();
        if parts.len() != 2 {
            return Err("Invalid target address format. Expected HOST:PORT".into());
        }
        (parts[0], parts[1].parse::<u16>()?)
    };

    tracing::info!("Starting port knocking sequence...");

    for knock_target in &cli.knock {
        let addr = format!("{}:{}", target_host, knock_target.port);
        
        match knock_target.protocol {
            Protocol::Tcp => {
                tracing::debug!("Knocking on TCP port {}", knock_target.port);
                // Use tokio's timeout for TCP knocking
                let connect_future = TcpStream::connect(&addr);
                if tokio::time::timeout(Duration::from_secs(1), connect_future).await.is_err() {
                    // We expect errors here, like connection refused or timeout.
                    // The goal is just to hit the port.
                }
            }
            Protocol::Udp => {
                tracing::debug!("Knocking on UDP port {}", knock_target.port);
                let socket = UdpSocket::bind("0.0.0.0:0").await?;
                socket.send_to(&[], &addr).await?;
            }
        }
    }

    tracing::debug!("Knocking sequence finished. Connecting to {}...", cli.target_address);

    let stream = TcpStream::connect(&cli.target_address).await?;

    tracing::info!("Connection established. Proxying data...");

    // Split the stream for bidirectional communication
    let (mut stream_read, mut stream_write) = stream.into_split();
    
    // Create stdin and stdout handles
    let mut stdin = tokio::io::stdin();
    let mut stdout = tokio::io::stdout();

    // Use tokio::select! to handle bidirectional data transfer
    tokio::select! {
        result = copy_with_flush(&mut stdin, &mut stream_write) => {
            if let Err(e) = result {
                tracing::error!("Error copying from stdin to stream: {}", e);
            }
        }
        result = copy_with_flush(&mut stream_read, &mut stdout) => {
            if let Err(e) = result {
                tracing::error!("Error copying from stream to stdout: {}", e);
            }
        }
    }
    
    Ok(())
}
